# -*- coding: utf-8 -*-

#  
#  Copyright 2014 Jérôme Kerdreux, Telecom Bretagne.
#  
#  This file is part of xAAL.
#  
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  		
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#  				
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#  


import serial
import sys,time,string


class X10Device:
    """
    Generic Mixing for all Devices :
    - Lamp
    - Switch
    """
    def __init__(self,controller,house,device ):
        """
        Constructor : 
         - controller : the CM11 Controller
         - house code : as string [A,F]
         - device code: as string [1,16]
         - device name: as string
        """
        self._controller = controller
        self.setHouseCode( house )
        self.setDeviceCode( device ) 
        
    def getController(self):
        return self._controller

    ### House Code ###
    def setHouseCode(self,house_code):
        self._houseCode = house_code
    
    def getHouseCode(self):
        return self._houseCode
    
    ### Device Code ###
    def setDeviceCode(self,device_code):
        self._deviceCode = str(device_code)

    def getDeviceCode(self):
        return self._deviceCode

    ### Sending ###
    def sendAction(self,actionName,value = 0 ):
        # sending address always start w/ 0x04 HeaderCode
        # cause we get 0000 0100
        controller = self.getController() 
        controller.send([0x04, getKeyForHouseAndDevice(self.getHouseCode() ,self.getDeviceCode()) ] )
        controller.isReady()
    
        # sending a action ( not a Dim ) always starts w/ 0x06 as header code 
        # cause we get 0000 0110
        dim = int( round( 22.0 * value ) / 100)
        header_code =   dim * 8  + 0x6
        print "Dim : %x : header : %x" % ( dim, header_code )
        controller.send([header_code, getKeyForHouseAndFunction(self.getHouseCode(), actionName ) ] )
        controller.isReady()

    def getTypeAsString(self):
        if isinstance(self,Switch):
            return 'Switch'
        elif isinstance(self,Lamp):
            return 'Lamp'





class Switch(X10Device):
    """ X10 Switch Device """

    def on(self):
        """ power on the switch """
        self.sendAction('On')

    def off(self):
        """ power off the switch """
        self.sendAction('Off')


class Lamp(X10Device):
    """ X10 Lamp Device """

    def on(self):
        self.sendAction('On')

    def off(self):
        self.sendAction('Off')

    def dim(self,value):
        self.sendAction('Dim', value)


class CM11Error(Exception):pass
class CM11:
    """ CM11 X10 Controller """

    def __init__(self):
        self._serial_port = None 

    def open(self,serial_port):
        """
        - Open the serial line with : 4800 N 1
        - set timeout to 5 sec
          ( pay caution that the device take ~ 4 sec to become ready after a cmd )
        serial_port = 0 => /dev/ttyS0 - COM1 on win32 
        """
        self.ser = serial.Serial(serial_port,baudrate = 4800, bytesize = 8, stopbits=1,parity='N',timeout=5)
        self._serial_port = serial_port
        self.setup()

    def getSerialPort(self):
        return self._serial_port
       
    def send(self,datas):
        """
        - send command to the CM11 controler
        - verify compute checksum
        """
        checksum = 0 
        # send data and compute checksum
        for data in datas :
            self.ser.write ( chr (data) )
        checksum = self.checksumForDatas( datas ) 

        # read the result 
        result = self.read()
        if result == None :
            raise CM11Error,'TimeOut in command %s' % str(datas)

        if result != checksum:
            raise CM11Error,"Checksum Error : wanted: %x / received: %x" %  (checksum,result)
        print "Checksum: %x" % checksum 


    def checksumForDatas(self,datas):
        """
        compute the checksum for a given datas
        take care that i found nowhere in doc
        why i should not use 0x9B for checksum
        computation
        """
        checksum = 0
        if datas[0] == 0x9B:
            datas = datas[1:]
        for data in datas :
            checksum = checksum + data
            if checksum > 0xFF : checksum = checksum - 256
        return checksum

    def isReady(self):
        """
        Send a Ok ? packet to check that device is ready
        ( Take care that this commit function call. You
          need to call this after each functions )
        """
        # sending back a 0x00 to OK the transmission
        self.ser.write( chr(0x00) )
        result = self.read()
        if result == None : 
            raise CM11Error,"TimeOut, while waiting for interface to be ready"
        elif result != 0x55:
            raise CM11Error,"Wrong ack: Interface isn't ready"

    def read(self):
        data = self.ser.read()
        if len(data) == 0 :
            return None # we found a timeout
        else:
            return ord( data )

    def setup(self):
        """
        look for a "Poll form interface"
        if this occur send a PC ready
        """
        # check for powserfail
        # wait until we receive a 0xa5
        print "Setting up the controller"
        data = self.read() 
        if data == 0xa5:
            print "We need to update the clock"
            self.setClock()

    def setClock(self):
        """ Update the CM11 controler internal clock """
        buf =[] 
        for i in range(0,7):
            buf.append('-')
        (year,month,day,hour,min,sec,week_day,year_day,dst) = time.localtime()

        buf[0] = 0x9b
        buf[1] = sec
        buf[2] = min + ( (hour %2) * 60 )      # min 0 - 119
        buf[3] = hour / 2                      # hour/2 0-11
        buf[4] = year_day  % 256               # convert the "int" to its byte value 
        buf[5] = ((year_day / 256) << 7)        # same 
        buf[5] |= ( 0x01 << week_day )         # bits 0-6 =  single bit mask day
        buf[6] = 0x06 << 4                     # dummy house code 
        buf[6] |= 0 ;
        self.send(buf)


### X10 helpers
def getKeyForHouseAndDevice(house_code,device_code):
    try:
        return HOUSE_CODES[house_code] * 0x10 + DEVICE_CODES[device_code]
    except KeyError:
        raise X10Error,"Wrong device adress '%s%s'" % ( house_code,device_code)


def getKeyForHouseAndFunction(house_code,function_name):
    try:
        return HOUSE_CODES[house_code] * 0x10 + FUNCTION_CODES[function_name]
    except KeyError:
        raise X10Error,"Wrong function %s on house %s" % (function_name,house_code)



HOUSE_CODES =  { 'A' : 0x06 , #0110
                 'B' : 0x0E , #1110
                 'C' : 0x02 , #0010
                 'D' : 0x0A , #1010
                 'E' : 0x01 , #0001
                 'F' : 0x09 , #1001
                 'G' : 0x05 , #0101
                 'H' : 0x0D , #1101
                 'I' : 0x07 , #0111
                 'J' : 0x0F , #1111
                 'K' : 0x03 , #0011
                 'L' : 0x0B , #1011
                 'M' : 0x00 , #0000
                 'N' : 0x08 , #1000
                 'O' : 0x04 , #0100
                 'P' : 0x0C , #1100
               }


DEVICE_CODES = { '1'  : 0x06 , #0110
                 '2'  : 0x0E , #1110
                 '3'  : 0x02 , #0010
                 '4'  : 0x0A , #1010
                 '5'  : 0x01 , #0001
                 '6'  : 0x09 , #1001
                 '7'  : 0x05 , #0101
                 '8'  : 0x0D , #1101
                 '9'  : 0x07 , #0111
                 '10' : 0x0F , #1111
                 '11' : 0x03 , #0011
                 '12' : 0x0B , #1011
                 '13' : 0x00 , #0000
                 '14' : 0x08 , #1000
                 '15' : 0x04 , #0100
                 '16' : 0x0C , #1100
                 }

# functions values ( this is linear but i do this to have the same API 
FUNCTION_CODES = { 'All Units Off'          : 0x00 , 
                   'All Lights On'          : 0x01 , 
                   'On'                     : 0x02 , 
                   'Off'                    : 0x03 , 
                   'Dim'                    : 0x04 , 
                   'Bright'                 : 0x05 ,
                   'All Lights Off'         : 0x06 ,
                   'Extended Code'          : 0x07 ,
                   'Hail Request'           : 0x08 ,
                   'Hail Acknoledge'        : 0x09 ,
                   'Pre-set Dim_1'          : 0x0A ,
                   'Pre-set Dim_2'          : 0x0B ,
                   'Extended Data Transfer' : 0x0C ,
                   'Status On'              : 0x0D ,
                   'Status Off'             : 0x0E ,
                   'Status Request'         : 0x0F ,
                 }

