# -*- coding: utf-8 -*-

#  
#  Copyright 2014 Jérôme Kerdreux, Philippe Tanguy Telecom Bretagne.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  		
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  				
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <http://www.gnu.org/licenses/>.
#  


import logging
import sys
import os
import time
import ConfigParser

from collections import deque

#Zwave
from libopenzwave import PyOptions
from libopenzwave import PyManager

#xAAL
from xAAL.Devices import XAALGateway 
from xAAL.Core import XAALEngine
from xAAL import tools

# xAAL-Zwave device
from Devices.fibaro.Fgs211 import Fgs211
from Devices.zwave_me.ZmeItemp import ZmeItemp
from Devices.aeon_labs.AeoHem3 import AeoHem3


class ZwaveController():
    
    def __init__(self,device='/dev/ttyUSB0',manager=None):
        self.zwTransceiver = device
        self.__manager = manager
        self._homeId = None
        self._nodeId = None

    def setHomeId(self,homeid=None):
        self._homeId = homeid 
    
    def setNodeId(self,nodeid=None):
        self._nodeId = nodeid 

    def start(self):
        self.__manager.addDriver(self.zwTransceiver)
   
    def stop(self):
            self.__manager.removeDriver(self.zwTransceiver)
            print "The Zwave driver has been removed"

    def info(self):
        print "---Zwave software---"
        print "Python library version: %s" % self.__manager.getPythonLibraryVersion()
        print "Openzwave library version: %s" % self.__manager.getOzwLibraryVersionNumber()
        print "Network home id : %s" % self._homeId
        print "Network node id : %s" % self.__manager.getControllerNodeId(self._homeId)
        print "Controller path : %s" % self.__manager.getControllerPath(self._homeId)
        print "Z-Wave API library version used by the controller : %s" % self.__manager.getLibraryVersion(self._homeId)
        print "Z-Wave library type used by controller : %s" % self.__manager.getLibraryTypeName(self._homeId)

    def getDriverStat(self):
        return self.__manager.getDriverStatistics(self._homeId)

    #-------------------
    # Manage Zwave network 
    #------------------
    def hardReset(self):
        print "-------------------------"
        print " ZWave hardware reset    "
        print "-------------------------"
        self.__manager.ResetController(self._homeId)
    
    def softReset(self):
        print "-------------------------"
        print " ZWave soft reset        "
        print "-------------------------"
        self.__manager.SoftReset(self._homeId)
    
    def testZwaveNetwork(self):
        self.__manager.testNetwork(self._homeId,count=3)
   
    def testZwaveNode(self,nodeId=None):
        self.__manager.testNetworkNode(self._homeId,count=3)
    
    def requestNodeState(self,nodeId=None):
        state = self.__manager.requestNodeState(self._homeId,nodeId)
        if state:
            print "Request sent with success to nodeId %s" % nodeId
            return True
        else:
            print "Request failed to nodeId %s" % nodeId
            return False


class ZwaveGateway(XAALGateway):

    def __init__(self,address=None,device=None):
        XAALGateway.__init__(self,address=address)
        self.setType('gateway.zwave')
        self.setVendorID("IHSEV")
        self.setProductID("Gateway ZWave")
       
        # Zwave
        self.networkState = None
        self.zwFifo = deque()
        self.options = PyOptions()
        self.configPath = self.options.getConfigPath()
        self.userPath = os.getcwd()
        self.options.create(self.configPath,self.userPath,"--logging false")
        self.options.addOptionBool("SaveConfiguration", False)
        self.options.lock()
        self.__zwManager = PyManager() 
        self.__zwManager.create()
        self.__zwManager.addWatcher(self.zwCallBack)
        self.__zwController = ZwaveController(device=device,manager=self.__zwManager)

    def getZwManager(self):
        return self.__zwManager

    #-------------------
    # Manage gateway
    #-------------------

    def start(self):
        print "-----------------------------"
        print "   Start Gateway xAAL-ZWave  "
        print "-----------------------------"
        self.__zwController.start()
        
        # Waiting for driver ready state
        timeout = 60*2
        while not self.networkState == 'STARTED':
            msg = self.readFifo()
            if msg:
                if msg['notificationType'] == 'DriverReady':
                    self.networkState = 'STARTED'
                    self.__zwController.setHomeId(msg['homeId'])
                    self.__zwController.setNodeId(msg['nodeId'])

            if timeout < 0:
                print "There is a problem with the driver:"
                print "- Verify if the controller is well plug"
                print "- Verify configuration"
                exit(0)
            
            time.sleep(1)
            timeout = timeout -1

        print "The driver is ready!"
        
        print "Initialization phase of the network, it can takes time, so be patient..."
        # Todo: consider network ready notification 
        while not self.networkState == 'AWAKED' or self.networkState=='READY':
            msgRx = self.readFifo()
            if msgRx:
                if msgRx['notificationType'] == 'AwakeNodesQueried':
                    self.networkState = 'AWAKED'
                elif msgRx['notificationType'] == 'AllNodesQueried':
                    self.networkState = 'READY'
                else:
                    self.forwardZwaveMsg(msg=msgRx)
        
        print "Awaked nodes queried, some sleeping nodes have not been queried but can exist"
        
    def stop(self):
        if not self.networkState == "STOPPED":
            self.__zwManager.removeWatcher(self.zwCallBack)
            self.__zwController.stop()
            self.networkState = "STOPPED"
            print "The Zwave network controller is now stopped"
        else:
            print "The Zwave network controller has been already stopped"

    def info(self):
        print "--------------------"
        print " Gateway xAAL-ZWave "
        print "--------------------"
        print "---Hardware---"
        print "Zwave transceiver: %s" % self.__zwController.zwTransceiver
        self.__zwController.info()
        print "---xAAL software---"
        devices = self.getDevices()
        print "Number of xAAL devices: %s" % len(devices)
        print "---xAAL devices list---"
        for device in devices:
            print "---"
            print "xAAL addr: %s" % device.getAddress()
            print "dev type: %s" % device.getType()

    def driverStat(self):
        stat = self.__zwController.getDriverStat()
        print "-------------------------"
        print " ZWave driver statistics "
        print "-------------------------"
        for key,val in stat:
            print "%s: %s" % (key,val)

    #-------------------
    # Manage xAAL msg 
    #------------------
    
    def xaalCallBack(self,msg,dev):
        r = None
        r = tools.runActions(msg,dev)
        if r:
            return r
    
    #-------------------
    # Manage zwave msg 
    #-------------------
    
    def zwCallBack(self,args):
        self.writeFifo(args)
        #print "[%s]" % args['notificationType'] 
        #print "- Node id: %s" % args['nodeId']
        
    def writeFifo(self,data):
        self.zwFifo.append(data)

    def readFifo(self):
        if self.zwFifo:
            data = self.zwFifo.popleft()
        else:
            data = None

        return data
    
    def loop(self):
            msgRx = self.readFifo()
            if msgRx:
                self.forwardZwaveMsg(msg=msgRx)

    def forwardZwaveMsg(self,msg=None):
        for device in self.getDevices():
            if msg['nodeId'] == device.getNodeId():
                device.parseMsg(msg)


def loadConfig():
    configFile = tools.getConfigFile(ZwaveGateway.__name__)
    cfg = tools.loadConfigFile(configFile)

    if cfg == None:
        cfg = tools.createConfigFile(configFile)
        cfg.add_section('gateway')
        cfg.set('gateway','xaaladdr',tools.getRandomUUID())
        cfg.set('gateway','serialport',"/dev/ttyUSB0")
        f = open(configFile,'w')
        cfg.write(f)

    xaalAddr = cfg.get('gateway','xaaladdr')
    port = cfg.get('gateway','serialport')

    gw=ZwaveGateway(address=xaalAddr,device=port)       

    for section in cfg.sections():
        dev = None
        if not section == 'gateway':
            xaalAddr = cfg.get(section,'xaaladdr')
            devType = cfg.get(section,'devtype')
            homeId = cfg.getint(section,'homeid')
            nodeId = cfg.getint(section,'nodeid')
           
            # Todo: modifier pour rendre plus generique car dependant des import ZmeItemp, Fgs211, ... 
            #donc si on ajout un device au .cfg il faut ajouter un import du device...
            if devType == 'temp.experimental':
                dev = ZmeItemp(xaalAddr=xaalAddr,
                        devtype=devType,
                        manager=gw.getZwManager(),
                        homeId=homeId,
                        nodeId=nodeId)
            elif devType == 'powerrelay.basic':
                dev = Fgs211(xaalAddr=xaalAddr,
                        devtype=devType,
                        manager=gw.getZwManager(),
                        homeId=homeId,
                        nodeId=nodeId)
            elif devType == 'energymeter.experimental':
                dev = AeoHem3(xaalAddr=xaalAddr,
                        devtype=devType,
                        manager=gw.getZwManager(),
                        homeId=homeId,
                        nodeId=nodeId)
            else:
                dev = None
        
        if dev:
            gw.addDevice(dev)
    
    return gw

if __name__ == '__main__':
    try:
        gw = loadConfig() 
        eng=XAALEngine()
        eng.setDevices([gw,] + gw.getDevices())
        eng.registerCallBackHandler(gw.xaalCallBack)
        gw.start()
        gw.info()
        eng.start()
        while 1:
            eng.loop()
            gw.loop()
    except KeyboardInterrupt:
        gw.stop()
        print "The Zwave Gateway is now closed"


