# -*- coding: utf-8 -*-

#  
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy, Telecom Bretagne.
#  
#  This file is part of xAAL.
# 
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#		
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#				
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#


from Network import NetworkConnector
from Messages import MessageFactory,MessageFactoryParserError
from Devices import XAALDeviceError
import tools
import os
import config


class XAALEngine:

    def __init__(self):
        self.__nc = NetworkConnector()
        self.__mf = MessageFactory()
        self.__devices = [] 
        self.__started=False
        self.__run=True
        self.__callbacks=[] 

    #####################################################
    # Attributes 
    #####################################################    
    # network connector 
    def getNetworkConnector(self):
        return self.__nc
    
    def setNetworkConnector(self,nc):
        self.__nc = nc            
    
    # message factory 
    def getMessageFactory(self):
        return self.__mf 

    def setMessageFactory(self,mf):
        self.__mf = mf

    # devices 
    def getDevices(self):
        return self.__devices

    def setDevices(self,devs):
        self.__devices = devs
        for i in self.__devices:
            i.setEngine(self)

    def registerDevice(self,dev):
        self.__devices.append(dev)
        dev.setEngine(self)

    def unregisterDevice(self,dev):
        dev.setEngine(None)        
        self.__devices.remove(dev)

        
    #####################################################
    # Message handling
    #####################################################    
    def sendMSG(self,msg):
        self.getNetworkConnector().send(msg)
        
    def buildMSG(self,dev=None,target=config.XAAL_BCAST_ADDR,msgtype=None,action=None,body=None):
        msg=self.getMessageFactory().buildMSG(dev,target,msgtype,action,body)
        return msg       

    def send(self,dev=None,target=config.XAAL_BCAST_ADDR,msgtype=None,action=None,body=None):
        msg = self.buildMSG(dev,target,msgtype,action,body)
        self.sendMSG(msg)

    def sendReply(self,dev,msg,body=None):
        self.send(dev=dev,target=msg.getSource(),msgtype='reply',action=msg.getAction(),body=body)


    def sendAlives(self):
        mf = self.getMessageFactory()
        for dev in self.getDevices():
            msg = mf.buildAlivefor(dev)
            self.sendMSG(msg)
    
    def error(self,dev,errcode,description=None):
        mf=self.getMessageFactory()
        msg=mf.buildErrorMSG(dev,errcode,description)
        self.sendMSG(msg)
    
    def attributesChange(self,dev,var):
        mf=self.getMessageFactory()
        body={var.getName():var.getValue()}
        msg=mf.buildMSG(dev,config.XAAL_BCAST_ADDR,"notify","attributesChange",body)
        self.sendMSG(msg)

 
    #####################################################
    # Callback
    #####################################################    
    # callback handler
    def registerCallBackHandler(self,func):
        self.__callbacks.append(func)

    def getCallBackHandlers(self):
        return self.__callbacks

    def unregisterCallBackHandler(self,func):
        self.getCallBackHandlers().remove(func)

    #####################################################
    # Bus Config
    #####################################################      
     
    def setBusConfig(self,addr,port,ttl):
        self.stop()
        nw=self.getNetworkConnector()
        nw.disconnect()
        nw.connect(addr,port,ttl)
        print "/!\ Warning /!\ BUS Config not SAVED ! To Be Done ! "
    
    def getBusConfig(self):
        return {"address" : self.getNetworkConnector().getAddr(), "port" : self.getNetworkConnector().getPort(), "ttl": self.getNetworkConnector().getTTL(),"stack_version:": self.getNetworkConnector().getStackVersion()}


    #####################################################
    # xAAL Handle request
    #####################################################
    def handleGetBusConfigRequest(self,msg,target):
        """ answer to getBusConfig"""
        self.sendReply(target,msg,body=self.getBusConfig())

    def handlerSetBusConfigRequest(self,msg,target):
        """ answer to SetBusconfig"""
        if target.getParent():
            self.error(target,500,"unable to change bus config for this device")

        body=msg.getBody()
        if 'addr'in body and 'port' in body and 'ttl' in body:
            self.setBusConfig(body['addr'], body['port'], body['ttl'])

    def handleCallBackRequest(self,msg,target):
        """ loop thought CallBackHandlers to execute the right action"""
        for handler in self.getCallBackHandlers():
            res=handler(msg,target)
            if res:
                self.sendReply(target,msg,body=res)


    #####################################################
    # Mainloops & run ..
    #####################################################
    def loop(self):
        """ process incomming data packets, and execute """

        # To Be Done methods 
        TBD = ['setCiphers','setKey','getCiphers']

        data = self.getNetworkConnector().getData()
        if data:
            try:
                msg = self.getMessageFactory().parseMSG(data)
            except MessageFactoryParserError:
                print "wrong msg drop"
            else:
                target=tools.filterMsgForDevices(msg,self.getDevices())
                if target:
                    for targ in target:
                        if msg.getMsgtype() == 'request':
                            action = msg.getAction()
                            if action =='getBusConfig': self.handleGetBusConfigRequest(msg,targ)
                            elif action =='setBusConfig': self.handlerSetBusConfigRequest(msg,targ)
                            elif action in TBD : self.error(targ,404,"%s Not implemented yet" % action)
                    
                        self.handleCallBackRequest(msg,targ)

        # timer handling .. 
        if self.__aliveTimer.reached():
            self.sendAlives()
            self.__aliveTimer.start()
    
    def start(self):
        self.sendAlives()
        self.__aliveTimer = tools.Timer(config.ALIVE_TIMER)
        self.__aliveTimer.start()
        self.__started = True
        
    def stop(self):
        self.__run=False
        
    def run(self):
        self.__run=True
        if not self.__started:
            self.start()
        
        while self.__run:
            self.loop()
                
