# -*- coding: utf-8 -*-

#  
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy, Telecom Bretagne.
#  
#  This file is part of xAAL.
# 
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#		
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#				
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

import tools
import random


def xaalAttributesChange(func):
    def wrapper(arg1,arg2):
        func(arg1,arg2)
        dev=arg1.getDevice()
        if dev:
            dev.attributesChange(arg1)
    return wrapper


class XAALAttribute:

    def __init__(self,name,dev,default=None):
        self.__name=name 
        self.__default=default
        self.__value=''
        self.__device=dev
        dev.addAttribute(self)
    
    #value 
    @xaalAttributesChange
    def setValue(self,value):
        self.__value=value
    
    def getValue(self):
        return self.__value

    #name
    def getName(self):
        return self.__name
        
    #default
    def getDefault(self):
        return self.__default
        
    #Device
    def setDevice(self,dev):
        self.__device=dev
        
    def getDevice(self):
        return self.__device


class XAALDeviceError(Exception):pass
class XAALDevice(): 

    def __init__(self,devtype=None,addr=None,parent=None,engine=None,): 
        self.setType(devtype)
        self.setAddress(addr)
        if parent:
            self.setParent(parent)
        else: self.__parent = None
        self.setEngine(engine)
        self.__vendorID= ""
        self.__productID= ""
        self.setVersion('Device version')
        self.__attributes=[]
        self.__actions = tools.findActions(self)    
    
    #####################################################
    # Attributes 
    #####################################################
    # Devtype
    def setType(self,value) :
        self.__devtype=value
        
    def getType(self):
        return self.__devtype                     
    
    # Address             
    def setAddress(self,value):
        if value==0:
            raise XAALDeviceError,"This address is reserved"
        self.__address=value
    
    def getAddress(self):
        return self.__address

    # Parent     
    def setParent(self,value):
        if isinstance(self,XAALGateway):
            raise XAALDeviceError,"Unable to set parent for Gateway"
        self.__parent = value
        
    def getParent(self):
        return self.__parent

    # Version
    def setVersion(self,value):
        self.__version=value
    
    def getVersion(self):
        return self.__version

    #vendorID
    def setVendorID(self,value):
        self.__vendorID=value
        
    def getVendorID(self):
        return self.__vendorID
        
    #productID
    def setProductID(self,value):
        self.__productID=value
   
    def getProductID(self):
        return self.__productID
        
    #attributes
    def addAttribute(self,value):
        self.__attributes.append(value) 
        
    def _getAttributes(self):
        return self.__attributes

    def attributesChange(self,var):     
        if self.getEngine():
            self.getEngine().attributesChange(self,var)

    #engine
    def setEngine(self,eng):
        self.__engine=eng

    def getEngine(self):
        return self.__engine        
        
    #####################################################
    # Properties 
    #####################################################
    devtype=property(getType,setType)
    address=property(getAddress,setAddress)
    parent=property(getParent,setParent)
    version=property(getVersion,setVersion)
       
    #####################################################
    # Usefull methods 
    #####################################################
    def dump(self):
        print "========= %s =========" % self
        print "Type:\t\t%s" % self.devtype
        if self.address:
            print "Address: \t0x%x" % self.address
        print "Version: \t%s" % self.version
        print "Parent: \t%s" % self.parent
        if isinstance(self,XAALGateway):
            print "Gateway for: \t%s" % self.devices
        if isinstance(self,XAALComposite):
            print "Composite for: \t%s" % self.devices
            print "Visibility: \t%s" % self.visibility

    #####################################################
    # Actions
    #####################################################
    # Return bound method
    def getActionsMethod(self):
        return self.__actions        
    
    # only Actions name
    def getActions(self):
        actions = {"actions": [action.__name__ for action in self.__actions]}
        return actions       
    getActions.exposed=True    
        
    def getDescription(self):
        result={}
        result['vendorID']=self.getVendorID()
        result['productID']=self.getProductID()
        result['version']=self.getVersion()
        if self.getParent():
            result['parent']=self.getParent().getAddress()
        return result
    getDescription.exposed=True
    
    def isAlive(self):
        return {"devtype":self.getType(),"address":self.getAddress()}
    isAlive.exposed=True    
        
    def getAttributes(self):
        result={}
        for attr in self._getAttributes():
            result.update({attr.getName():attr.getValue()})
        return result
    getAttributes.exposed=True


class XAALVirtual(XAALDevice):
   
    def __init__(self,devtype=None,address=None,parent=None):
        XAALDevice.__init__(self,devtype,address,parent)
        self.__devtype='virtual.experimental'


class _XAALParent(XAALDevice):
    
    def __init__(self,devtype=None,address=None,parent=None):
        self.__devices=[]        
        XAALDevice.__init__(self,devtype,address,parent)
        
    # devices    
    def addDevice(self,dev):
        if isinstance(dev,XAALGateway):
            if self.getEngine():
                self.getEngine().error(self,2)
            raise XAALGatewayError,"unable to attach a Gateway to anything"
            
        if isinstance(self,XAALComposite):
            if isinstance(dev,XAALComposite):           
                if self.getEngine():
                   self.getEngine().error(self,3)
                raise XAALCompositeError,"unable to attach a Composite to another one"

        self.__devices.append(dev)
        dev.setParent(self)
        
    def removeDevice(self,dev):   
        self.__devices.remove(dev)
        dev.setParent(None)
        
    def setDevices(self,dev):
        self.__devices=dev
        
    def getDevices(self):        
         return self.__devices    
         
    def getDevicesDesc(self):
        res={}
        typedev=""
        n=1
        for i in self.__devices:
            
            if i.getType() in res:
                typedev=i.getType()+str(n)
            else:
                typedev=i.getType()
                
            res[typedev]=i.getAddress()
            n+=1
        return res
    getDevicesDesc.exposed=True     
    devices=property(getDevices,setDevices)
    
    
class XAALCompositeError(Exception):pass
class XAALComposite(_XAALParent): 
     
    def __init__(self,address=None,parent=None):
        self.__visibility=True        
        _XAALParent.__init__(self,None,address,parent)
        self.setType('composite.experimental')
        
    def setVisibility(self,value):
        if value==True:
            self.__visibility = True
        else:
            self.__visibility = False
 
    def getVisibility(self):
        return self.__visibility

    visibility=property(getVisibility,setVisibility)


class XAALGatewayError(Exception):pass         
class XAALGateway(_XAALParent):

    def __init__(self,address=None):
        _XAALParent.__init__(self,None,address,None)
        self.setType('gateway.experimental')




